#!/usr/bin/env python
"""
usage: managed.py <index>

where <index> is a key in the locmap dictionary

For new indexes, you will need to add them to locmap, loccols, and the
appropriate index_* dict (depending on whether the index is the subdirectory of
a build or not). Then, add the tool_data_table_conf.xml entry to
/cvmfs/data.galaxyproject.org/managed/location/tool_data_table_conf.xml
"""

import glob
import os
import os.path
import shutil
import subprocess
import sys
import tempfile

user = 'g2test'
#server = 'login1.corral.tacc.utexas.edu'
#remote_base = '/corral-repl/tacc/galaxy'
#remote_data = '/corral-repl/tacc/galaxy/manageddata/data'
server = 'galaxy07.tacc.utexas.edu'
remote_base = '/galaxy-repl'
remote_data = '/galaxy-repl/manageddata/data'
local_base = '/cvmfs/data.galaxyproject.org/managed'
rewrite_base = '/galaxy-repl/manageddata/data'

locmap = {
    #'dbkeys': ['test/tool_data/toolshed.g2.bx.psu.edu/repos/devteam/data_manager_fetch_genome_dbkeys_all_fasta/776bb1b478a0/all_fasta.loc',
    #           'test/tool_data/toolshed.g2.bx.psu.edu/repos/devteam/data_manager_fetch_genome_dbkeys_all_fasta/776bb1b478a0/dbkeys.loc'],
    #'all_fasta': 'test/tool_data/toolshed.g2.bx.psu.edu/repos/devteam/data_manager_fetch_genome_all_fasta/2ebc856bce29/all_fasta.loc',
    'dbkeys': 'test/tool_data/toolshed.g2.bx.psu.edu/repos/devteam/data_manager_fetch_genome_dbkeys_all_fasta/776bb1b478a0/dbkeys.loc',
    'all_fasta': ['test/tool_data/toolshed.g2.bx.psu.edu/repos/devteam/data_manager_fetch_genome_all_fasta/2ebc856bce29/all_fasta.loc',
                  'test/tool_data/toolshed.g2.bx.psu.edu/repos/devteam/data_manager_fetch_genome_dbkeys_all_fasta/776bb1b478a0/all_fasta.loc'],
    'twobit': ['test/tool_data/toolshed.g2.bx.psu.edu/repos/devteam/data_manager_twobit_builder/74b09c8e5f6e/alignseq.loc',
               'test/tool_data/toolshed.g2.bx.psu.edu/repos/devteam/data_manager_twobit_builder/74b09c8e5f6e/twobit.loc',
               'test/tool_data/toolshed.g2.bx.psu.edu/repos/devteam/data_manager_twobit_builder/74b09c8e5f6e/lastz_seqs.loc'],
    'hisat2': 'test/tool_data/toolshed.g2.bx.psu.edu/repos/iuc/data_manager_hisat2_index_builder/98a60a4cfb9a/hisat2_indexes.loc',
    'bowtie2': ['test/tool_data/toolshed.g2.bx.psu.edu/repos/devteam/data_manager_bowtie2_index_builder/e87aeff2cf88/bowtie2_indices.loc',
                'test/tool_data/toolshed.g2.bx.psu.edu/repos/devteam/data_manager_bowtie2_index_builder/e87aeff2cf88/tophat2_indices.loc'],
    'sam': 'test/tool_data/toolshed.g2.bx.psu.edu/repos/devteam/data_manager_sam_fasta_index_builder/cf875cbe2df4/fasta_indexes.loc',
    'picard': 'test/tool_data/toolshed.g2.bx.psu.edu/repos/devteam/data_manager_picard_index_builder/00491eabe22b/picard_index.loc',
    'plant_tribes_scaffolds': 'test/tool_data/toolshed.g2.bx.psu.edu/repos/iuc/data_manager_plant_tribes_scaffolds_downloader/13c235ded82e/plant_tribes_scaffolds.loc',
    'bwa_mem': 'test/tool_data/toolshed.g2.bx.psu.edu/repos/devteam/data_manager_bwa_mem_index_builder/cb0147ade868/bwa_mem_index.loc',
    'kraken': 'test/tool_data/testtoolshed.g2.bx.psu.edu/repos/devteam/data_manager_build_kraken_database/d5d803e654dd/kraken_databases.loc',
    'taxonomy': 'test/tool_data/toolshed.g2.bx.psu.edu/repos/devteam/data_manager_fetch_ncbi_taxonomy/fafcd2f5db36/ncbi_taxonomy.loc',
    'rnastar': 'test/tool_data/toolshed.g2.bx.psu.edu/repos/iuc/data_manager_star_index_builder/cdc4d8a998e1/rnastar_index2.loc',
    'maf_indexes': 'test/tool_data/toolshed.g2.bx.psu.edu/repos/dave/data_manager_fetch_index_maf/edf39ed96bc3/maf_indexes.loc',
    'rnastar_index2x_versioned': 'test/tool_data/toolshed.g2.bx.psu.edu/repos/iuc/data_manager_star_index_builder/6c6c6df09e64/rnastar_index2x_versioned.loc',
    'kraken2_databases': 'test/tool_data/toolshed.g2.bx.psu.edu/repos/iuc/data_manager_build_kraken2_database/0eebe086fd58/kraken2_databases.loc',
    'selection_background': 'test/tool_data/toolshed.g2.bx.psu.edu/repos/iuc/data_manager_selection_background/e4f07770b340/selection_background.loc',
}

loccols = {
    #'dbkeys': [
    #    {
    #        'build': 0,
    #        'dbkey': 1,
    #        'path': 3,
    #    },
    #    {
    #        'dbkey': 0,
    #        'path': 2,
    #    },
    #],
    #'all_fasta': {
    #    'build': 0,
    #    'dbkey': 1,
    #    'path': 3,
    #},
    'dbkeys': {
        'dbkey': 0,
        'path': 2,
    },
    'all_fasta': [
        {
            'build': 0,
            'dbkey': 1,
            'path': 3,
        },
        {
            'build': 0,
            'dbkey': 1,
            'path': 3,
        },
    ],
    'twobit': [
        # FIXME: the field I'm calling dbkey is probably actually the build, but we don't have any "sub-build" builds in these DMs yet...
        {
            'dbkey': 1,
            'path': 2,
        },
        {
            'dbkey': 0,
            'path': 1,
        },
        {
            'dbkey': 0,
            'path': 2,
        },
    ],
    'hisat2': {
        'build': 0,
        'dbkey': 1,
        'path': 3,
    },
    'bowtie2': [
        {
            'build': 0,
            'dbkey': 1,
            'path': 3,
        },
        {
            'build': 0,
            'dbkey': 1,
            'path': 3,
        },
    ],
    'sam': {
        'build': 0,
        'dbkey': 1,
        'path': 3,
    },
    'picard': {
        'build': 0,
        'dbkey': 1,
        'path': 3,
    },
    'bwa_mem': {
        'build': 0,
        'dbkey': 1,
        'path': 3,
    },
    'kraken': {
        'path': 2,
    },
    'taxonomy': {
        'path': 2,
    },
    'rnastar': {
        'build': 0,
        'dbkey': 1,
        'path': 3,
    },
    'plant_tribes_scaffolds': {
        'build': 0,
        'dbkey': 1,
        'path': 2,
    },
    'maf_indexes': {
        'name': 0,
        'dbkey': 1,
        'indexed_for': 2,
        'exists_in_maf': 3,
        'path': 4,
    },
    'rnastar_index2x_versioned': {
        'value': 0,
        'dbkey': 1,
        'name': 2,
        'path': 3,
        'with_gene_model': 4,
        'version': 5,
    },
    'kraken2_databases': {
        'value': 0,
        'name': 1,
        'path': 2,
    },
    'selection_background': {
        'value': 0,
        'label': 1,
        'path': 2,
    },
}

index_subdirs = {
    'dbkeys': 'len',
    'all_fasta': 'seq',
    'twobit': 'seq',
    'hisat2': 'hisat2_index',
    'bowtie2': 'bowtie2_index',
    'sam': 'sam_indexes',
    'picard': 'picard_index',
    'bwa_mem': 'bwa_mem_index',
    'rnastar': 'rnastar_index2',
    'maf_indexes': 'align',
}

index_dirs = {
    'kraken2_databases': 'kraken2_databases',
    'kraken': 'kraken_database',
    'taxonomy': 'ncbi_taxonomy',
    'plant_tribes_scaffolds': 'plant_tribes',
    'rnastar_index2x_versioned': 'rnastar',
    'selection_background': 'selection_background',
}


# overrides filenames for overlapping tables from different DMs
#loc_file_map = {
#    'dbkeys': [ 'all_fasta_dbkeys.loc', 'dbkeys.loc' ],
#}
loc_file_map = {
    'all_fasta': [ 'all_fasta.loc', 'all_fasta_dbkeys.loc' ],
}


rsync_excludes = {
    'dbkeys': ('len/ucsc/', 'len/compute')
}


def _listify(obj):
    if not isinstance(obj, list):
        return [obj]
    return obj


def _local_loc(key, i, remote_loc_rel_path):
    if key in loc_file_map:
        local_loc_rel_path = _listify(loc_file_map[key])[i]
    else:
        local_loc_rel_path = os.path.basename(remote_loc_rel_path)
    return os.path.join(local_base, 'location', local_loc_rel_path)


def get_loc(index):
    for i, remote_loc_rel_path in enumerate(_listify(locmap[index])):
        remote = user + '@' + server + ':' + os.path.join(remote_base, remote_loc_rel_path)
        scp = ['scp', remote, _local_loc(index, i, remote_loc_rel_path)]
        #print 'Executing: %s' % ' '.join(scp)
        subprocess.check_call(scp)


def fix_loc(index):
    for i, remote_loc_rel_path in enumerate(_listify(locmap[index])):
        with tempfile.NamedTemporaryFile() as new:
            loc_path = _local_loc(index, i, remote_loc_rel_path)
            with open(loc_path, 'r') as old:
                print 'transmogrify %s' % loc_path
                for line in old:
                    if line.startswith('#'):
                        new.write(line)
                        continue
                    if not line.strip():
                        continue
                    cols = line.split('\t')
                    loc_cols = _listify(loccols[index])[i]
                    assert cols[loc_cols['path']].startswith(rewrite_base), 'line: %s; loc_cols: %s, cols: %s' % (line, loc_cols, cols)
                    if index in index_subdirs:
                        rewrite = os.path.join(rewrite_base, cols[loc_cols['dbkey']])
                    else:
                        rewrite = rewrite_base
                    cols[loc_cols['path']] = cols[loc_cols['path']].replace(rewrite, local_base)
                    new.write('\t'.join(cols))
            new.flush()
            shutil.copy(new.name, loc_path)


def fix_loc_perms(index):
    for i, remote_loc_rel_path in enumerate(_listify(locmap[index])):
        os.chmod(_local_loc(index, i, remote_loc_rel_path), 0644)


def rsync_data(index):
    if index in index_subdirs:
        remote_path = os.path.join(remote_data, '*', index_subdirs[index])
    else:
        remote_path = os.path.join(remote_data, index_dirs[index])
    if rsync_excludes.get(index):
        excludes = ['--exclude=' + p for p in rsync_excludes[index]]
    else:
        excludes = []
    rsync = ['rsync', '-avP', '--delete', '--exclude=.cvmfscatalog'] + excludes + [user + '@' + server + ':' + remote_path, local_base]
    subprocess.check_call(rsync)


def fix_seq_links(index):
    # TODO: if any of the index_dirs tools need fasta fixing you'll have to address that here
    if index in index_dirs:
        return
    fastas = os.path.join(local_base, index_subdirs[index], '*', '*.fa')
    for link in glob.glob(fastas):
        target = os.readlink(link)
        if os.path.islink(link) and target.startswith('../../../../../../galaxy/data/'):
            # lies outside managed data (so it's by hand data)
            print 'Symlink %s target was: %s' % (link, target),
            target = target.replace('../../../../../../galaxy/data/', '../../../byhand/')
            os.unlink(link)
            os.symlink(target, link)
            print 'now: %s' % target
        assert os.path.exists(link), 'Bad symlink: %s' % link


def main():
    assert len(sys.argv) == 2, 'usage: managed.py <index>'
    index = sys.argv[1]
    assert index in locmap, 'error: %s not in locmap' % index
    assert index in index_subdirs.keys() + index_dirs.keys(), 'error: %s not in index_(sub)?dirs' % index
    rsync_data(index)
    fix_seq_links(index)
    get_loc(index)
    fix_loc(index)
    fix_loc_perms(index)

if __name__ == '__main__':
    main()
